==> netsa-io-silk_2.12-1.5.2-sources.jar: io/SnappyOutputStreamBuffer.scala <==
// Copyright 2015-2022 by Carnegie Mellon University
// See license information in LICENSE.txt

package org.cert.netsa.io.silk
package io

import java.io.{DataOutputStream, OutputStream}
import org.apache.hadoop.conf.Configuration
import org.apache.hadoop.io.compress.{CompressionCodecFactory, Compressor}


/**
  * A [[BufferWriter]] which compresses data using Snappy and writes
  * it to a stream in a form compatible with SiLK file streams, using
  * Hadoop compression codecs.
  *
  * @see [[SnappyOutputStreamBuffer$ the companion object]] for more
  * details.
  */
private[silk] class SnappyOutputStreamBuffer(
  configuration: Configuration,
  outStream:     OutputStream)
    extends BufferWriter
{
  // The output stream
  private val out: DataOutputStream = outStream match {
    case out: DataOutputStream => out
    case _ => new DataOutputStream(outStream)
  }

  // The compression engine
  private val compressor: Compressor = {
    val cf = new CompressionCodecFactory(configuration)
    val c = cf.getCodecByName("snappy")
    c.createCompressor()
  }

  /**
    * Return the size of a buffer that is guaranteed to be large
    * enough to hold the result of compressing `sourceLen` bytes of
    * data.
    */
  private[this] def compressBound(sourceLen: Int): Int = {
    // Formula taken from body of the MaxCompressedLength function in
    // the snappy 1.1.3 source code.
    return 32 + sourceLen + sourceLen/6;
  }

  /**
    * Compress and write any pending data, finalize the compression
    * engine, and flush the output stream.
    */
  def end(): Unit = {
    assert(compressor.needsInput)
    //println("end called")
    compressor.end()
    out.flush()
  }

  /**
    * Write the first `length` bytes of data in `buffer` to the output
    * stream.
    *
    * @throws java.io.IOException when an error occurs.
    */
  def putBuffer(uncompressedData: Array[Byte], uncompressedSize: Int): Unit = {
    if (0 == uncompressedSize) {
      return
    }

    // create buffer to hold the result.  FIXME: Consider moving this
    // array outside the function
    val maxCompressedSize = compressBound(uncompressedSize)
    val compressedData = Array.ofDim[Byte](maxCompressedSize)

    //println("putBuffer called with buffer size = " + uncompressedSize +
    //  "; maxCompressedSize = " + maxCompressedSize)

    // compress it
    assert(compressor.needsInput)
    compressor.setInput(uncompressedData, 0, uncompressedSize)
    compressor.finish()

    val compressedSize = compressor.compress(compressedData, 0, maxCompressedSize)

    //println("Actual compressed size is " + compressedSize)
    //println("total uncompressed bytes: " + compressor.getBytesRead() +
    //  "; total compressed bytes: " + compressor.getBytesWritten())
    //println("Compressor needsInput is " + compressor.needsInput +
    //  " and finished is " + compressor.finished())

    // write the two sizes and the compressed block
    out.writeInt(compressedSize)
    out.writeInt(uncompressedSize)
    out.write(compressedData, 0, compressedSize)

    // the following block should not be necessary but it is here
    // "just in case"
    while ( !compressor.finished() ) {
      val size = compressor.compress(compressedData, 0, maxCompressedSize)
      //println("Writing an additional block of " + size + " bytes")
      out.write(compressedData, 0, size)
    }

    assert(compressor.needsInput)
    assert(0 == compressor.compress(compressedData, 0, maxCompressedSize))
    compressor.reset()
  }
}


/**
  * The SnappyOutputStreamBuffer object provides support for creating an
  * [[SnappyOutputStreamBuffer]].
  */
private[silk] object SnappyOutputStreamBuffer {
  def apply(conf: Configuration, outStream: OutputStream):
      SnappyOutputStreamBuffer =
  {
    new SnappyOutputStreamBuffer(conf, outStream)
  }
}

// @LICENSE_FOOTER@
//
// Copyright 2015-2022 Carnegie Mellon University. All Rights Reserved.
//
// This material is based upon work funded and supported by the
// Department of Defense and Department of Homeland Security under
// Contract No. FA8702-15-D-0002 with Carnegie Mellon University for the
// operation of the Software Engineering Institute, a federally funded
// research and development center sponsored by the United States
// Department of Defense. The U.S. Government has license rights in this
// software pursuant to DFARS 252.227.7014.
//
// NO WARRANTY. THIS CARNEGIE MELLON UNIVERSITY AND SOFTWARE ENGINEERING
// INSTITUTE MATERIAL IS FURNISHED ON AN "AS-IS" BASIS. CARNEGIE MELLON
// UNIVERSITY MAKES NO WARRANTIES OF ANY KIND, EITHER EXPRESSED OR
// IMPLIED, AS TO ANY MATTER INCLUDING, BUT NOT LIMITED TO, WARRANTY OF
// FITNESS FOR PURPOSE OR MERCHANTABILITY, EXCLUSIVITY, OR RESULTS
// OBTAINED FROM USE OF THE MATERIAL. CARNEGIE MELLON UNIVERSITY DOES NOT
// MAKE ANY WARRANTY OF ANY KIND WITH RESPECT TO FREEDOM FROM PATENT,
// TRADEMARK, OR COPYRIGHT INFRINGEMENT.
//
// Released under a GNU GPL 2.0-style license, please see LICENSE.txt or
// contact permission@sei.cmu.edu for full terms.
//
// [DISTRIBUTION STATEMENT A] This material has been approved for public
// release and unlimited distribution. Please see Copyright notice for
// non-US Government use and distribution.
//
// Carnegie Mellon(R) and CERT(R) are registered in the U.S. Patent and
// Trademark Office by Carnegie Mellon University.
//
// This software includes and/or makes use of third party software each
// subject to its own license as detailed in LICENSE-thirdparty.tx
//
// DM20-1143

==> netsa-io-silk_2.12-1.6.0-sources.jar: io/SnappyOutputStreamBuffer.scala <==
// Copyright 2015-2022 by Carnegie Mellon University
// See license information in LICENSE.txt

package org.cert.netsa.io.silk
package io

import java.io.{DataOutputStream, OutputStream}
import org.xerial.snappy.Snappy

/**
  * A [[BufferWriter]] which compresses data using Snappy and writes
  * it to a stream in a form compatible with SiLK file streams, using
  * Hadoop compression codecs.
  *
  * @see [[SnappyOutputStreamBuffer$ the companion object]] for more
  * details.
  */
private[silk] class SnappyOutputStreamBuffer(
  outStream:     OutputStream)
    extends BufferWriter
{
  // The output stream
  private val out: DataOutputStream = outStream match {
    case out: DataOutputStream => out
    case _ => new DataOutputStream(outStream)
  }

  /** Finish writing compressed data and flush output. */
  def end(): Unit = {
    out.flush()
  }

  /** Write the first `length` bytes of data in `buffer` to the output
    * stream.
    *
    * @throws java.io.IOException when an error occurs.
    */
  def putBuffer(uncompressedData: Array[Byte], uncompressedSize: Int): Unit =
    if ( uncompressedSize > 0 ) {
      // create buffer to hold the result.  FIXME: Consider moving this
      // array outside the function
      val maxCompressedSize = Snappy.maxCompressedLength(uncompressedSize)
      val compressedData = Array.ofDim[Byte](maxCompressedSize)

      val compressedSize = Snappy.compress(
        uncompressedData, 0, uncompressedSize, compressedData, 0)

      // write the two sizes and the compressed block
      out.writeInt(compressedSize)
      out.writeInt(uncompressedSize)
      out.write(compressedData, 0, compressedSize)
    }
}


/**
  * The SnappyOutputStreamBuffer object provides support for creating an
  * [[SnappyOutputStreamBuffer]].
  */
private[silk] object SnappyOutputStreamBuffer {
  def apply(outStream: OutputStream): SnappyOutputStreamBuffer =
    new SnappyOutputStreamBuffer(outStream)
}

// @LICENSE_FOOTER@
//
// Copyright 2015-2022 Carnegie Mellon University. All Rights Reserved.
//
// This material is based upon work funded and supported by the
// Department of Defense and Department of Homeland Security under
// Contract No. FA8702-15-D-0002 with Carnegie Mellon University for the
// operation of the Software Engineering Institute, a federally funded
// research and development center sponsored by the United States
// Department of Defense. The U.S. Government has license rights in this
// software pursuant to DFARS 252.227.7014.
//
// NO WARRANTY. THIS CARNEGIE MELLON UNIVERSITY AND SOFTWARE ENGINEERING
// INSTITUTE MATERIAL IS FURNISHED ON AN "AS-IS" BASIS. CARNEGIE MELLON
// UNIVERSITY MAKES NO WARRANTIES OF ANY KIND, EITHER EXPRESSED OR
// IMPLIED, AS TO ANY MATTER INCLUDING, BUT NOT LIMITED TO, WARRANTY OF
// FITNESS FOR PURPOSE OR MERCHANTABILITY, EXCLUSIVITY, OR RESULTS
// OBTAINED FROM USE OF THE MATERIAL. CARNEGIE MELLON UNIVERSITY DOES NOT
// MAKE ANY WARRANTY OF ANY KIND WITH RESPECT TO FREEDOM FROM PATENT,
// TRADEMARK, OR COPYRIGHT INFRINGEMENT.
//
// Released under a GNU GPL 2.0-style license, please see LICENSE.txt or
// contact permission@sei.cmu.edu for full terms.
//
// [DISTRIBUTION STATEMENT A] This material has been approved for public
// release and unlimited distribution. Please see Copyright notice for
// non-US Government use and distribution.
//
// Carnegie Mellon(R) and CERT(R) are registered in the U.S. Patent and
// Trademark Office by Carnegie Mellon University.
//
// This software includes and/or makes use of third party software each
// subject to its own license as detailed in LICENSE-thirdparty.tx
//
// DM20-1143
